from collections import defaultdict
from functools import reduce, partial
from operator import getitem
from typing import Optional

from flatten_dict import unflatten

from chaiverse.lib.dict_tools import deep_get
from chaiverse.database import firebase_database as firebase_db


class MockDatabase():
    def __init__(self):
        nested_dict = lambda: defaultdict(nested_dict)
        self.store = nested_dict()

    def get(self, path: str, shallow=False):
        keys = self._get_keys(path)
        record = self._get_from_store(keys) or {}
        if shallow:
            record = {key: True for key, _ in record.items()}
        record = None if record in [{}, []] else record
        return record

    def is_in_database(self, path: str):
        keys = self._get_keys(path)
        is_in = self._is_in_store(keys)
        record = None
        if is_in:
            record = self.get(path)
        return record is not None

    def set(self, path: str, value: dict):
        keys = self._get_keys(path)
        value = firebase_db.serialise_input(value)
        value = dict_to_default_dict(value)
        self._set_in_store(keys, value)

    def update(self, path: str, record: dict):
        keys = self._get_keys(path)
        record = firebase_db.serialise_input(record)
        record = dict_to_default_dict(record)
        for record_keys, value in record.items():
            self._update_in_store(keys + record_keys.split('/'), value)

    def multi_update(self, path: str, record: dict):
        record = unflatten(record, splitter="path")
        self.update(path, record)

    def where(self, path, **kwargs):
        keys = self._get_keys(path)
        records = self._get_from_store(keys) or {}
        records = list(records.values())
        records = firebase_db.filter_records(records, kwargs)
        return records

    def _remove_recursively(self, keys, current_node, current_index):
        key = keys[current_index]
        if key in current_node:
            if current_index == len(keys) - 1:
                current_node.pop(key)
            else:
                self._remove_recursively(keys, current_node[key], current_index + 1)
                if len(current_node[key]) == 0:
                    current_node.pop(key)

    def remove(self, path: str):
        keys = self._get_keys(path)
        self._remove_recursively(keys, self.store, 0)

    def query_by_key_range(self, path, start_at=None, end_at=None, limit_to_first: Optional[int] = None, limit_to_last: Optional[int] = None):
        if limit_to_first and limit_to_last:
            raise ValueError('Cannot set both first and last limits.')
        keys = self._get_keys(path)
        records = self._get_from_store(keys) or {}
        record_tuples = list(records.items())
        key_func = lambda tuple: tuple[0]
        record_tuples = self._range_query_modifier(record_tuples, start_at, end_at, limit_to_first, limit_to_last, key_func=key_func)
        records = {key: value for key, value in record_tuples} or None
        return records

    def query_by_child_value_range(self, path, by, start_at=None, end_at=None, limit_to_first: Optional[int] = None, limit_to_last: Optional[int] = None):
        if limit_to_first and limit_to_last:
            raise ValueError('Cannot set both first and last limits.')
        keys = self._get_keys(path)
        records = self._get_from_store(keys) or {}
        record_tuples = list(records.items())
        key_func = lambda tuple: deep_get(tuple[1], by)
        record_tuples = self._range_query_modifier(record_tuples, start_at, end_at, limit_to_first, limit_to_last, key_func=key_func)
        records = {key: value for key, value in record_tuples} or None
        return records

    def _range_query_modifier(self, record_tuples, start_at, end_at, limit_to_first, limit_to_last, key_func):
        record_tuples = sorted(record_tuples, key=key_func)
        record_tuples = [
            record_tuple for record_tuple in record_tuples
            if (not start_at or key_func(record_tuple) >= start_at) and (not end_at or key_func(record_tuple) <= end_at)
        ]
        if limit_to_first:
            record_tuples = record_tuples[:limit_to_first]
        if limit_to_last:
            record_tuples = record_tuples[-limit_to_last:]
        return record_tuples

    def atomic_add(self, path: str, value: float):
        func = partial(firebase_db.add, y=value)
        self.atomic_set(path, func)

    def atomic_increment(self, path: str):
        self.atomic_set(path, firebase_db.incrementer)

    def atomic_decrement(self, path: str):
        self.atomic_set(path, firebase_db.decrementer)

    def atomic_set(self, path, operation: callable):
        value = self.get(path)
        value = operation(value)
        self.set(path, value)

    @staticmethod
    def _get_keys(path: str):
        path = path.lstrip("/")
        path = path.rstrip("/")
        keys = path.split("/")
        # To avoid using emptystring as key when
        # interfacing with root leaf
        keys = keys if keys != [""] else []
        return keys

    def _get_from_store(self, keys):
        node = self.store
        for key in keys:
            if key not in node:
                node = None
                break
            node = node[key]
        return node

    def _is_in_store(self, keys):
        node = self.store
        is_in = True
        for key in keys:
            if key not in node:
                is_in = False
                break
            node = node[key]
        return is_in

    def _get_or_create_from_store(self, keys):
        return reduce(getitem, keys, self.store) if keys else self.store

    def _set_in_store(self, keys, value):
        _raise_for_disallowed_key(keys)
        _raise_for_disallowed_key(value)
        value = _deep_remove_none(value) if isinstance(value, dict) else value
        if keys:
            self._get_or_create_from_store(keys[:-1])[keys[-1]] = value
        else:
            self.store = value

    def _update_in_store(self, keys, value):
        current_value = self._get_or_create_from_store(keys[:-1]) or {}
        current_value = current_value.get(keys[-1])
        # Updating is only relevant if both stored value and new value are
        # dicts
        if isinstance(current_value, dict) and isinstance(value, dict):
            value = deep_update(current_value, value)
        self._set_in_store(keys, value)

    @property
    def nested_default_dict(self):
        # To make it easy to do self.store["some"]["key"] without raising
        # KeyError
        return defaultdict(self.nested_default_dict)


def _get_disallowed_firebase_keys():
    return [".", "$", "#", "[", "]", "/"]


def _raise_for_disallowed_key(obj):
    disallowed_keys = _get_disallowed_firebase_keys()
    if isinstance(obj, str):
        msg = f"Key {obj} contains disallowed character!"
        assert not any(disallowed in obj for disallowed in disallowed_keys), msg
    elif isinstance(obj, list):
        [_raise_for_disallowed_key(key) for key in obj]
    elif isinstance(obj, dict):
        [_raise_for_disallowed_key(key) for key in obj.keys()]
        [
            _raise_for_disallowed_key(value) for value in obj.values()
            if isinstance(value, dict)
        ]


def _string_is_disallowed(string):
    disallowed_keys = _get_disallowed_firebase_keys()
    return any(disallowed in string for disallowed in disallowed_keys)


def dict_to_default_dict(d):
    if isinstance(d, dict):
        nested_dict = lambda: defaultdict(nested_dict)
        d = defaultdict(nested_dict, {k: dict_to_default_dict(v) for k, v in d.items()})
    return d


def deep_update(mapping, updating_mapping, traverse_dict=True):
    mapping = mapping.copy()
    mapping = _deep_update(mapping, updating_mapping, traverse_dict=traverse_dict) if len(updating_mapping) else {}
    return mapping


def _deep_update(mapping, updating_mapping, traverse_dict=True):
    for k, v in updating_mapping.items():
        if k in mapping and isinstance(mapping[k], dict) and isinstance(v, dict) and traverse_dict:
            # Firebase does not support updating deeply nested dicts, so we
            # ensure that if we go any deeper, we are not updating by setting
            # traverse_dict = False
            mapping[k] = deep_update(mapping[k], v, traverse_dict=False)
        else:
            mapping[k] = v
    return mapping


def _deep_remove_none(mapping):
    for key, value in mapping.items():
        value = _deep_remove_none(value) if isinstance(value, dict) else value
    # We must do it this way to preserve the infinite dict structure of the
    # mapping
    none_keys = [key for key, value in mapping.items() if value == None]
    for key in none_keys:
        mapping.pop(key)
    return mapping



from chaiverse.http_client import AuthClient


def get_login_url(redirect_uri, provider="huggingface"):
    auth_client = AuthClient()
    payload = {"redirect_uri": redirect_uri}
    response = auth_client.post(f"/oauth2/{provider}/authorize", json=payload)
    return response["login_url"]



from datetime import datetime
import json
from math import inf
from typing import List, Optional, Union, Dict

import numpy as np
from pydantic import BaseModel, Extra, Field, model_validator
from typing_extensions import Literal

from chaiverse import config
from chaiverse.lib.now import utcnow
from chaiverse.lib.pydantic_tools import get_fields_in_schema, UnionParser
from chaiverse.lib.date_tools import convert_to_us_pacific_date


class BaseLeaderboardRow(BaseModel, extra=Extra.allow):
    developer_uid: str
    submission_id: str
    model_name: Optional[str] = ""
    model_group: Optional[str] = ""
    status: str
    timestamp: datetime = Field(default_factory=utcnow)

    # preference metrics
    num_battles: int = 0
    num_wins: int = 0
    celo_rating: Optional[float] = None

    # family friendly scores
    family_friendly_score: float = 0
    family_friendly_standard_error: float | None = None

    @model_validator(mode="after")
    def validate_model_group(self):
        if not self.model_group:
            model_repo = getattr(self, "model_repo", "")
            self.model_group = model_repo[:24]
        return self

    @property
    def display_name(self):
        name = self.model_name if self.model_name else self.submission_id
        return name

    @property
    def win_ratio(self):
        return self.num_wins / self.num_battles if self.num_battles > 0 else None

    @property
    def us_pacific_date(self):
        us_pacific_date = convert_to_us_pacific_date(self.timestamp).date()
        return us_pacific_date

    @property
    def ranking_group(self):
        ''' this is a string that labels whether this type of model will be ranked in single or blended group '''
        return None

    @property
    def is_internal_developer(self):
        is_internal = self.developer_uid in config.INTERNAL_USERS
        return is_internal

    @property
    def ineligible_reason(self) -> Optional[str]:
        return None

    def can_auto_deactivate(self):
        is_deployed = self.status == 'deployed'
        has_enough_battles = self.num_battles >= config.AUTO_DEACTIVATION_MIN_NUM_BATTLES
        return is_deployed and has_enough_battles

    def all_fields_dict(self):
        fields = get_fields_in_schema(self.__class__)
        fields_dict = {key: getattr(self, key) for key in fields}
        return fields_dict

    def base_fields_dict(self):
        fields = BaseLeaderboardRow.__fields__.keys()
        fields_dict = {key: getattr(self, key) for key in fields}
        return fields_dict

    ## pydantic 2.x implements model_dump. Existing dict() in pydantic 1.x doesn't serialize recursively
    #def model_dump(self):
    #    return json.loads(self.json())


class SingleGroupLeaderboardRow(BaseLeaderboardRow):
    ''' minimum implementation to show a model in single-model leaderboard '''
    submission_type: str = Field(default='single_model')

    @property
    def ranking_group(self):
        return 'single'


class BlendedGroupLeaderboardRow(BaseLeaderboardRow):
    ''' minimum implementation for a model in blended leaderboard '''
    submission_type: Literal['blended_model'] = Field(default='blended_model')

    @property
    def ranking_group(self):
        return 'blended'


class LatencyRecord(BaseModel):
    batch_size: int
    throughput: float  # throughput for all the samples provided, subject to window dressing error
    latency_mean: float  # a better approximate of throughput without window dressing error
    latency_p50: float  # a better typical latency estimate removing outliners
    latency_p90: float  # proxy for worst case latency estiamte


class BasicLeaderboardRow(SingleGroupLeaderboardRow):
    submission_type: Literal['basic'] = Field(default='basic')
    model_repo: str
    model_architecture: Optional[str] = ""
    model_num_parameters: float | None = None
    best_of: int = 1
    max_input_tokens: int
    max_output_tokens: Optional[int] = 64
    reward_model: Optional[str] = "default"

    # throughput and cost metrics
    latencies: Optional[List[LatencyRecord]] = Field(default=None)
    gpu_counts: Optional[Dict[str, int]] = Field(default=None)

    @property
    def language_model(self):
        return self.model_repo

    @property
    def model_size(self):
        size_gb = round(self.model_num_parameters / 1e9) if self.model_num_parameters else None
        size_gb = f'{size_gb}B'
        return size_gb

    @property
    def ineligible_reason(self) -> Optional[str]:
        reason = None
        if self.status != 'deployed' and self.num_battles < config.LEADERBOARD_STABLE_ELO_REQUIRED_BATTLES:
            reason = f'num_battles<{config.LEADERBOARD_STABLE_ELO_REQUIRED_BATTLES}'
        if self.max_output_tokens != config.DEFAULT_MAX_OUTPUT_TOKENS and self.max_output_tokens != None:
            reason = f'max_output_tokens!={config.DEFAULT_MAX_OUTPUT_TOKENS}'
        if self.status not in ['inactive', 'deployed', 'torndown']:
            reason = 'model is not deployable'
        if self.developer_uid == config.E2E_DEVELOPER_UID:
            reason = 'model is only for e2e test'
        return reason

    @property
    def throughput_3p7s(self):
        return self.throughput(3.7)

    def throughput(self, target_latency):
        return get_throughput(target_latency, self.latencies)


def get_throughput(target_latency, records: Optional[List[LatencyRecord]]):
    throughput = None
    if records:
        latency_means = [record.latency_mean for record in records]
        batch_sizes = [record.batch_size for record in records]
        if min(latency_means) <= target_latency <= max(latency_means):
            target_batch_size = np.interp(target_latency, latency_means, batch_sizes)
            throughput = target_batch_size / target_latency
            throughput = round(throughput, 2)
            throughput = _convert_nan_to_none(throughput)
    return throughput


def _convert_nan_to_none(value):
    value = None if np.isnan(value) else value
    return value


class FunctionLeaderboardRow(SingleGroupLeaderboardRow):
    submission_type: Literal['function'] = Field(default='function')


class BlendLeaderboardRow(BlendedGroupLeaderboardRow):
    submission_type: Literal['blend'] = Field(default='blend')
    submissions: List[str]

    @property
    def language_model(self):
        return ','.join(self.submissions)

    @property
    def model_size(self):
        return 'n/a'


class TaggedSubmissionID(BaseModel):
    submission_id: str
    tags: Optional[List[str]] = None


LeaderboardRow = Union[
    BasicLeaderboardRow,
    FunctionLeaderboardRow,
    BlendLeaderboardRow,
    SingleGroupLeaderboardRow,  # simplest implementation for a single model leaderboard row
    BlendedGroupLeaderboardRow,  # simplest implementation for a blended model leaderboard row
]



import glob
import os
import uuid

import asgi_correlation_id
from prometheus_client import multiprocess, CollectorRegistry
from prometheus_fastapi_instrumentator import Instrumentator, metrics


class FastAPIInstrumentator():
    def __init__(self, fastapi_app, config: "ChaiverseServerConfig"):
        self.fastapi_app = fastapi_app
        self.config = config

    def instrument(self):
        if self._should_instrument():
            self._instrument_prometheus()
            self._patch_prometheus_middleware()
            self._instrument_tracing()

    def _instrument_prometheus(self):
        registry = CollectorRegistry()
        instrumentator = Instrumentator(registry=registry)
        instrumentator.add(
            metrics.default(
                custom_labels=self.config.prometheus_labels,
                metric_subsystem=self.config.name,
                latency_lowr_buckets=self.config.prometheus_latency_buckets,
                registry=registry
            )
        )
        instrumentator.instrument(self.fastapi_app)
        instrumentator.expose(self.fastapi_app)

    def _instrument_tracing(self):
        self.fastapi_app.add_middleware(
            asgi_correlation_id.CorrelationIdMiddleware,
            header_name=self.config.trace_header_name,
            validator=None,
            transformer=_transform_trace_id,
        )

    def _patch_prometheus_middleware(self):
        # Change the behaviour of the instrumentation middleware so that it
        # returns fully interpolated paths in the prometheus labels instead of
        # leaving them uninterpolated
        middleware = [
            middleware for middleware in self.fastapi_app.user_middleware
            if "Prometheus" in middleware.cls.__name__
        ]
        middleware = middleware[0]
        def _get_handler(self, request):
            return request.url.path, True
        middleware.cls._get_handler = _get_handler

    def _should_instrument(self):
        # This avoids a bug on later version of fastapi where it crashes
        # if you try to add a middleware more than once (useful in testing)
        should_add_middleware = True
        if self.fastapi_app.middleware_stack is not None:
            should_add_middleware = False
        return should_add_middleware


def _transform_trace_id(trace_id):
    # https://cloud.google.com/trace/docs/trace-context
    parts = trace_id.split('-')
    if len(parts) > 1:
        trace_id = parts[1]
    return trace_id
